import subprocess
import os
import sys
import copy
from pathlib import Path
from datetime import datetime
import numpy as np
from scipy.io import loadmat
import h5py


__all__ = [
    "Reader",
    "Writer",
    "convert_li_to_h5",
]


THIS_DIRECTORY = os.path.abspath(os.path.dirname(__file__))


class Reader(object):
    """
    File reader for splendaq HDF5 files.

    Attributes
    ----------
    filename : str
        The path to the HDF5 file to open.

    """

    def __init__(self, filename=None):
        """
        Initialization of the HDF5 file reader.

        Parameters
        ----------
        filename : str, optional
            The path to the HDF5 file to open.

        """

        self.filename = filename

    def _open_file(self, include_data, include_metadata, filename=None):
        """
        Hidden method to handle opening data under different settings.

        """

        if filename is not None:
            self.filename = filename
        elif self.filename is None:
            raise ValueError("No filename specified.")

        with h5py.File(self.filename, mode='r') as hf:
            if include_metadata:
                out_dict = dict(hf.attrs)
                for key in hf:
                    if key != 'data':
                        out_dict[key] = np.asarray(hf[key])
            if include_data:
                data = np.asarray(hf['data'])

        if include_data and include_metadata:
            return data, out_dict
        elif include_data and not include_metadata:
            return data
        elif not include_data and include_metadata:
            return out_dict


    def get_data(self, filename=None, include_metadata=False):
        """
        Method to load the data from the specified HDF5 file.

        Parameters
        ----------
        filename : str, optional
            The path to the HDF5 file to open. Can be set here if not
            set when initializing, or if a different file will be read.
        include_metadata : bool, optional
            If True, the metadata will also be returned alongside the
            data in the HDF5 file. Default is False.

        Returns
        -------
        data : ndarray
            The array of trace data saved to the HDF5 file.
        out_dict : dict, optional
            A dictionary containing the corresponding metadata for the
            events.

        """

        return self._open_file(True, include_metadata, filename=filename)

    def get_metadata(self, filename=None):
        """
        Method to load only the metadata from the specified HDF5 file.

        Parameters
        ----------
        filename : str, optional
            The path to the HDF5 file to open. Can be set here if not
            set when initializing, or if a different file will be read.

        Returns
        -------
        out_dict : dict
            A dictionary containing the corresponding metadata for the
            events.

        """

        return self._open_file(False, True, filename=filename)


class Writer(object):
    """
    File writer for splendaq HDF5 files.

    Attributes
    ----------
    filename : str
        The path and filename to save the HDF5 file as.

    """

    def __init__(self, filename=None):
        """
        Initialization of the HDF5 file reader.

        Parameters
        ----------
        filename : str, optional
            The path to the HDF5 file to open.

        """

        self.filename = filename

    def write_data(self, data, filename=None, **metadata):
        """
        Method to save the data to the specified HDF5 file.

        Parameters
        ----------
        data : ndarray
            The ndarray of data to save to the HDF5 file.
        filename : str, optional
            The path to the HDF5 file to open. Can be set here if not
            set when initializing, or if a different file will be read.
        metadata : kwargs
            All of the corresponding metadata for the events in the
            HDF5 file.

        """

        current_metadata_size = 0

        if filename is not None:
            self.filename = filename
        elif self.filename is None:
            raise ValueError("No filename specified.")

        with h5py.File(self.filename, mode='w') as hf:
            hf.create_dataset(
                'data',
                data=data,
                compression='gzip',
                track_order=True,
            )
            for key in metadata:
                size_of_key = sys.getsizeof(copy.deepcopy(metadata[key]))
                if current_metadata_size + size_of_key < 64000:
                    current_metadata_size += size_of_key
                    hf.attrs[key] = metadata[key]
                else:
                    hf.create_dataset(
                        key, data=metadata[key], compression='gzip',
                    )
                


def convert_li(file, my_os, filetype='mat'):
    """
    Python wrapper for liconvert executable for converting `.li` files
    to `.csv`, `.mat`, or `.npy`.

    Parameters
    ----------
    file : str
        The absolute file path and name to the `.li` file to convert.
    my_os : str
        The operating system to assume when trying to convert a file.
        Can be 'mac', 'linux', or 'windows'.
    filetype : str, optional
        The file type to convert `file` to, can be "csv", "npy", or
        "mat". Default is "mat".

    Returns
    -------
    subprocess_obj : CompletedProcess
        The completed subprocess, including information on what was
        run and error code.

    Raises
    ------
    ValueError
        If `os` is not one of 'mac', 'linux', or 'windows'.

    """

    if my_os == "mac":
        exe = "liconvert_macos"
    elif my_os == "linux":
        exe = "liconvert_linux"
    elif my_os == "windows":
        exe = "liconvert_windows.exe"
    else:
        raise ValueError("Check docstring for supported OSs")

    return subprocess.run(
        [os.path.join(
            THIS_DIRECTORY,
            f'_liconvert{os.sep}{exe}',
        ), f"--{filetype}", file],
        stdout=False,
    )


def convert_li_to_h5(li_file, my_os):
    """
    Converting a LI file to a splendaq HDF5 file.

    Parameters
    ----------
    li_file : str
        The path to the binary LI file to convert to a splendaq HDF5
        file.
    my_os : str
        The operating system to assume when trying to convert a file.
        Can be 'mac', 'linux', or 'windows'.

    """

    convert_li(li_file, my_os, filetype='mat')
    mat_filename = Path(li_file).with_suffix('.mat')
    mat_file = loadmat(mat_filename, simplify_cells=True)['moku']
    os.remove(mat_filename)

    comment_str = mat_file['comment']
    comment = list(
        filter(None, comment_str.replace('\r\n', '').split('% '))
    )

    kw_timestamp = "Acquired"
    moku_timestamp = [
        s for s in comment if kw_timestamp in s
    ][0][len(kw_timestamp):].replace(' ','')
    iso_timestamp = moku_timestamp[:-2] + ':' + moku_timestamp[-2:]
    begintime = datetime.fromisoformat(iso_timestamp)
    seriesnumber = int(begintime.strftime("%y%m%d%H%M%S"))
    epochtime = begintime.timestamp()

    kw_fs = "Acquisition rate:"
    fs = float([
        s for s in comment if kw_fs in s
    ][0][len(kw_fs):].split('Hz')[0])

    arr = mat_file['data']
    columns = mat_file['legend']

    new_filename = Path(li_file).with_suffix('.h5')
    datashape = arr.T[None, 1:].shape

    FW = Writer(new_filename)
    FW.write_data(
        data=arr.T[None, 1:],
        channels=list(columns[1:]),
        comment=comment,
        fs=fs,
        datashape=datashape,
        eventindex=[0] * datashape[0],
        eventnumber=np.arange(datashape[0]),
        eventtime=[epochtime] * datashape[0],
        seriesnumber=[seriesnumber] * datashape[0],
        dumpnumber=[1] * datashape[0],
        triggertime=[epochtime] * datashape[0],
        triggertype=[0] * datashape[0],
        parentseriesnumber=[seriesnumber] * datashape[0],
        parenteventnumber=np.arange(datashape[0]),
    )




import numpy as np
import serial
import time


__all__ = [
    "SR560",
]


def wait(secs):
    def decorator(func):
        def wrapper(*args, **kwargs):
            ret = func(*args, **kwargs)
            time.sleep(secs)
            return ret
        return wrapper
    return decorator


class SR560(object):
    """
    Class for controlling an SRS SR560 low-noise preamplifier.
    Documentation on the various commands can be found in the SR560
    user manual.

    """

    def __init__(self, port):
        """
        Initialization of the SS560 class. Establishes a connection to
        a specified port.

        Parameters
        ----------
        port : str
            The port to connect to that the SR560 is on,
            e.g. "COM4".

        Returns
        -------
        None

        """

        self._SRS = serial.Serial()
        self._SRS.port = port
        self._SRS.open()

    @wait(1)
    def set_coupling(self, value):
        """
        Set the coupling of the SR560.

        Parameters
        ----------
        value : int
            Sets the input coupling. Coupling to ground is 0,
            DC coupling is 1, and AC coupling is 2.

        Returns
        -------
        None

        """

        self._SRS.write(f"CPLG{value}\r\n".encode())

    @wait(1)
    def set_dynamic_reserve(self, value):
        """
        Set the dynamic reserve of the SR560.

        Parameters
        ----------
        value : int
            Sets the dynamic reserve (DR). Low noise is 0,
            high DR is 1, and calibration gains is 2.

        Returns
        -------
        None

        """

        self._SRS.write(f"DYNR{value}\r\n".encode())

    @wait(1)
    def set_blanking(self, value):
        """
        Operates amplifier blanking.

        Parameters
        ----------
        value : int
            Set 0 to for not blanked. Set to 1 for blanked.

        Returns
        -------
        None

        """

        self._SRS.write(f"BLINK{value}\r\n".encode())

    @wait(1)
    def set_filter_mode(self, value):
        """
        Sets filter mode.

        Parameters
        ----------
        value : int
            The filter mode to set, values below
            0 = bypass
            1 =  6 dB lowpass
            2 = 12 dB lowpass
            3 =  6 dB highpass
            4 = 12 dB highpass
            5 = bandpass

        Returns
        -------
        None

        """

        self._SRS.write(f"FLTM{value}\r\n".encode())

    @wait(1)
    def set_gain(self, gain):
        """
        Controls the gain of the SR560.
        
        Parameters
        ----------
        gain : int
            Sets the gain of the device, allowed values are
            1 to 50000 in 1-2-5 sequence.

        Returns
        -------
        None

        """

        gain_dict = {
            1 : 0,
            2 : 1,
            5 : 2,
            10 : 3,
            20 : 4,
            50 : 5,
            100 : 6,
            200 : 7,
            500 : 8,
            1000 : 9,
            2000 : 10,
            5000 : 11,
            10000 : 12,
            20000 : 13,
            50000 : 14,
        }

        if gain not in gain_dict:
            raise ValueError(
                "Not a valid gain, see docstring for allowed values."
            )

        value = gain_dict[gain]
        self._SRS.write(f"GAIN{value}\r\n".encode())

    @wait(1)
    def set_highpass_filter(self, freq):
        """
        Set the highpass filter frequency.

        Parameters
        ----------
        freq : float
            The highpass filter frequency to be set. Defined in steps
            from 0.03, 0.1, 0.3, 1, ..., 10000 Hz.

        Returns
        -------
        None

        """

        freq_dict = {
            0.03 : 0,
            0.1 : 1,
            0.3 : 2,
            1 : 3,
            3 : 4,
            10 : 5,
            30 : 6,
            100 : 7,
            300 : 8,
            1000 : 9,
            3000 : 10,
            10000 : 11,
        }

        if freq not in freq_dict:
            raise ValueError(
                "Not a valid highpass freq, "
                "see docstring for allowed values."
            )

        value = freq_dict[freq]
        self._SRS.write(f"HFRQ{value}\r\n".encode())

    @wait(1)
    def set_signal_invert(self, value):
        """
        Sets the signal invert status.

        Parameters
        ----------
        value : int
            0 is noninverted. 1 is inverted.

        Returns
        -------
        None

        """

        self._SRS.write(f"INVT{value}\r\n".encode())

    @wait(1)
    def listen_all(self):
        """
        Makes all attached SR560s listeners.

        Returns
        -------
        None

        """

        self._SRS.write(b"LALL\r\n")

    @wait(1)
    def listen(self, value):
        """
        Listen to a specific SR560.

        Parameters
        ----------
        value : int
            Makes SR560 with address `value` a listener.
            Can be 0, 1, 2, or 3.

        Returns
        -------
        None

        """

        self._SRS.write(f"LISN{value}\r\n".encode())

    @wait(1)
    def set_lowpass_filter(self, freq):
        """
        Set the lowpass filter frequency.

        Parameters
        ----------
        freq : float
            The lowpass filter frequency to be set. Defined in steps
            from 0.03, 0.1, 0.3, 1, ..., 1000000 Hz.

        Returns
        -------
        None

        """

        freq_dict = {
            0.03 : 0,
            0.1 : 1,
            0.3 : 2,
            1 : 3,
            3 : 4,
            10 : 5,
            30 : 6,
            100 : 7,
            300 : 8,
            1000 : 9,
            3000 : 10,
            10000 : 11,
            30000 : 12,
            100000 : 13,
            300000 : 14,
            1000000 : 15,
        }

        if freq not in freq_dict:
            raise ValueError(
                "Not a valid lowpass freq, "
                "see docstring for allowed values."
            )

        value = freq_dict[freq]
        self._SRS.write(f"LFRQ{freq}\r\n".encode())

    @wait(1)
    def reset_overload(self):
        """
        Resets overload for 0.5 seconds.

        Returns
        -------
        None

        """

        self._SRS.write(b"ROLD\r\n")

    @wait(1)
    def set_input_source(self, value):
        """
        Sets the input source.

        Parameters
        ----------
        value : int
            Input source. Set 0 for "A",
            set to 1 for "A-B", and set to 2 for "B".

        Returns
        -------
        None

        """

        self._SRS.write(f"SRCE{value}\r\n".encode())

    @wait(1)
    def set_vernier_gain_status(self, value):
        """
        Set the vernier gain status.

        Parameters
        ----------
        value : int
            Set to 0 for cal'd gain. Set to 1 for vernier gain.

        Returns
        -------
        None

        """

        self._SRS.write(f"UCAL{value}\r\n".encode())

    @wait(1)
    def set_vernier_gain(self, gain):
        """
        Set the vernier gain on the SR560.

        Parameters
        ----------
        gain : float
            Set the vernier gain in 0.5% steps from 0 to 100.

        Returns
        -------
        None

        """
        
        allowed_vals = np.linspace(0, 100, num=201)

        if gain not in allowed_vals:
            raise ValueError(
                "Vernier gain must be between 0 and 100 in steps of 0.5."
            )

        self._SRS.write(f"UCGN{gain}\r\n".encode())

    @wait(1)
    def unlisten_all(self):
        """
        Unlisten. Unaddresses all attached SR560s.

        Returns
        -------
        None

        """

        self._SRS.write(b"UNLS\r\n")

    @wait(1)
    def reset_to_defaults(self):
        """
        Reset. Recalls default settings.

        Returns
        -------
        None

        """

        self._SRS.write(b"*RST\r\n")

    def close(self):
        """
        Close connection to the port to free it for other users.

        Returns
        -------
        None

        """
        
        self._SRS.close()



import numpy as np
from glob import glob
from pathlib import Path
from collections import Counter
from itertools import chain
import os
from datetime import datetime
from scipy.signal import correlate
import warnings

from splendaq.io import Reader, Writer


__all__ = [
    'EventBuilder',
]


def rand_sections(x, n, l):
    """
    Return random, non-overlapping sections of an n-dimensional array.
    For greater-than-1 dimensional arrays, the function finds random
    sections along the last axis.

    Parameters
    ----------
    x : ndarray
        n dimensional array to choose sections from.
    n : int
        Number of sections to choose.
    l : int
        Length in bins of sections.

    Returns
    -------
    evtinds : ndarray
        Array of the corresponding event indices for each section.
    res : ndarray
        Array of the n sections of x, each with length l.

    """

    t = np.arange(x.shape[0]) * x.shape[-1]

    tup = ((n,), x.shape[1:-1], (l,))
    sz = sum(tup, ())

    res = np.zeros(sz)
    evtinds = np.zeros(n, dtype=int)
    j = 0

    nmax = int(x.shape[-1] / l)

    if x.shape[0] * nmax < n:
        raise ValueError(
            "Either n or l is too large, trying to find more "
            "random sections than are possible."
        )

    choicelist = list(range(len(x))) * nmax
    rows = np.random.choice(choicelist, size=n, replace=False)
    counts = Counter(rows)


    for key in counts:
        offset = 0
        ncounts = counts[key]
        inds = np.arange(x.shape[-1] - (l - 1) * ncounts)

        for ind in sorted(np.random.choice(
            inds, size=ncounts, replace=False,
        )):
            ind += offset
            res[j] = x[key, ..., ind:ind + l]
            evtinds[j] = t[key] + ind + l//2
            j += 1
            offset += l - 1

    return evtinds, res


class EventBuilder(object):
    """
    Class to build events from continuous data, for both randomly
    triggered events and threshold triggered events.

    """

    def __init__(self, contdatadir, savepath, tracelength,
                 maxevtsperdump=500):
        """
        Initialization of the EventBuilder class.

        Parameters
        ----------
        contdatadir : str
            Absolute path to the directory containing the continuous
            data.
        savepath : str
            Absolute path to the directory to save the built events.
        tracelength : int
            The length of each event built in units of bins.
        maxevtsperdump : int, optional
            The maximum number of events to save to each file created.

        """

        self._contdatadir = contdatadir
        self._tracelength = tracelength
        self._maxevtsperdump = maxevtsperdump
        self._savepath = f"{os.path.abspath(savepath)}{os.sep}"

        self._start = datetime.now()

        filelist = sorted(glob(f"{self._contdatadir}/*.h5"))
        FR = Reader(filelist[0])
        metadata = FR.get_metadata()
        self._fs = metadata['fs']

        self._phi = None
        self._norm = None
        self._resolution = None

        self._template = None
        self._psd = None
        self._nthreshold_on = None
        self._nthreshold_off = None
        self._tchan = None

        self._threshold_on = None
        self._threshold_off = None


    def acquire_randoms(self, nrandoms):
        """
        Method for acquiring randomly triggered events and
        saving them to the specified file location.

        Parameters
        ----------
        nrandoms : int
            The number of randoms to acquire from the continuous
            data.

        """

        savename = "randoms_" + self._start.strftime("%Y%m%d_%H%M%S")
        seriesnumber = int(self._start.strftime("%y%m%d%H%M%S"))

        filelist = sorted(glob(f"{self._contdatadir}/*.h5"))
        datashapes = []
        for filename in filelist:
            FR = Reader(filename)
            metadata = FR.get_metadata()
            datashapes.append(metadata['datashape'])
        datashapes_arr = np.vstack(datashapes)

        nmax_rows = datashapes_arr[..., -1] // self._tracelength
        nrows = datashapes_arr[:, 0]

        arr_range = np.arange(len(filelist))

        choicelist = list(
            chain(
                *[[a] * n * nr for a, n, nr in zip(
                    arr_range, nmax_rows, nrows,
                )]
            )
        )

        if nrandoms > len(choicelist):
            warnings.warn(
                f"The desired number of randoms is too high, as there "
                f"{len(choicelist)} possible randoms. Defaulting to this "
                "maximum amount of randoms."
            )
        rows = np.random.choice(
            choicelist,
            size=np.min([nrandoms, len(choicelist)]),
            replace=False,
        )
        counts = Counter(rows)

        evtinds_list = []
        traces_list = []
        parentsn_list = []
        parenten_list = []
        epochtime_list = []

        evt_counter = 0
        dumpnum = 1
        basenevents = 0

        for key in sorted(counts.keys()):

            FR = Reader(filelist[key])
            data, metadata = FR.get_data(include_metadata=True)
            parentsn = metadata['parentseriesnumber'][0]
            parenten = metadata['parenteventnumber'][0]
            epochtime_start = metadata['eventtime'][0]

            ei, tr = rand_sections(data, counts[key], self._tracelength)

            evt_counter += len(ei)

            evtinds_list.append(ei)
            traces_list.append(tr)
            parentsn_list.append(parentsn * np.ones(len(ei)))
            parenten_list.append(parenten * np.ones(len(ei)))
            epochtime_list.append(epochtime_start * np.ones(len(ei)))

            if evt_counter >= self._maxevtsperdump:

                evtinds = np.concatenate(evtinds_list)
                traces = np.vstack(traces_list)
                parentsns = np.concatenate(parentsn_list)
                parentens = np.concatenate(parenten_list)
                epochtimes = np.concatenate(epochtime_list)

                del evtinds_list
                del traces_list
                del parentsn_list
                del parenten_list
                del epochtime_list

                for ii in range(len(evtinds) // self._maxevtsperdump):

                    nevents = len(evtinds[:self._maxevtsperdump])
                    FW = Writer(
                        f"{self._savepath}{savename}_F{dumpnum:04d}.h5",
                    )
                    FW.write_data(
                        data=traces[:self._maxevtsperdump],
                        eventindex=evtinds[:self._maxevtsperdump],
                        eventnumber=np.arange(nevents) + basenevents,
                        eventtime=(
                            epochtimes[:self._maxevtsperdump]
                        ) + (
                            evtinds[:self._maxevtsperdump] / self._fs
                        ),
                        seriesnumber=[seriesnumber] * nevents,
                        dumpnumber=[dumpnum] * nevents,
                        triggertime=np.zeros(nevents),
                        triggertype=np.zeros(nevents, dtype=int),
                        triggeramp=np.zeros(nevents),
                        parentseriesnumber=parentsns[:self._maxevtsperdump],
                        parenteventnumber=parentens[:self._maxevtsperdump],
                        datashape=traces[:self._maxevtsperdump].shape,
                        fs=self._fs,
                        channels=metadata['channels'],
                        comment='randoms',
                    )
                    dumpnum += 1
                    basenevents += nevents

                    evtinds = evtinds[self._maxevtsperdump:]
                    traces = traces[self._maxevtsperdump:]
                    parentsns = parentsns[self._maxevtsperdump:]
                    parentens = parentens[self._maxevtsperdump:]
                    epochtimes = epochtimes[self._maxevtsperdump:]

                if len(evtinds) > 0:
                    evtinds_list = [evtinds]
                    traces_list = [traces]
                    parentsn_list = [parentsns]
                    parenten_list = [parentens]
                    epochtime_list = [epochtimes]
                    evt_counter = len(evtinds)
                else:
                    evtinds_list = []
                    traces_list = []
                    parentsn_list = []
                    parenten_list = []
                    epochtime_list = []
                    evt_counter = 0

        # clean up the remaining events
        if evt_counter > 0:

            evtinds = np.concatenate(evtinds_list)
            traces = np.vstack(traces_list)
            parentsns = np.concatenate(parentsn_list)
            parentens = np.concatenate(parenten_list)
            epochtimes = np.concatenate(epochtime_list)

            del evtinds_list
            del traces_list
            del parentsn_list
            del parenten_list
            del epochtime_list

            dumps_left = len(evtinds) / self._maxevtsperdump

            for ii in range(np.ceil(dumps_left).astype(int)):

                nevents = len(evtinds[:self._maxevtsperdump])
                FW = Writer(f"{self._savepath}{savename}_F{dumpnum:04d}.h5")
                FW.write_data(
                    data=traces[:self._maxevtsperdump],
                    eventindex=evtinds[:self._maxevtsperdump],
                    eventnumber=np.arange(nevents) + basenevents,
                    eventtime=(
                        epochtimes[:self._maxevtsperdump]
                    ) + (
                        evtinds[:self._maxevtsperdump] / self._fs
                    ),
                    seriesnumber=[seriesnumber] * nevents,
                    dumpnumber=[dumpnum] * nevents,
                    triggertime=np.zeros(nevents),
                    triggertype=np.zeros(nevents, dtype=int),
                    triggeramp=np.zeros(nevents),
                    parentseriesnumber=parentsns[:self._maxevtsperdump],
                    parenteventnumber=parentens[:self._maxevtsperdump],
                    datashape=traces[:self._maxevtsperdump].shape,
                    fs=self._fs,
                    channels=metadata['channels'],
                    comment='randoms',
                )
                dumpnum += 1
                basenevents += nevents

                dumps_left = len(evtinds) / self._maxevtsperdump

                if ii + 1 != np.ceil(dumps_left).astype(int):
                    evtinds = evtinds[self._maxevtsperdump:]
                    traces = traces[self._maxevtsperdump:]
                    parentsns = parentsns[self._maxevtsperdump:]
                    parentens = parentens[self._maxevtsperdump:]
                    epochtimes = epochtimes[self._maxevtsperdump:]


    def _initialize_filter(self):
        """
        Method for initializing the optimal filter coefficients.

        """

        # calculate the time-domain optimum filter
        phi_freq = np.fft.fft(self._template)/self._psd
        phi_freq[0] = 0 # ensure we do not use DC information
        self._phi = np.fft.ifft(phi_freq).real
        # calculate the normalization of the optimum filter
        self._norm = np.dot(self._phi, self._template)
        
        # calculate the expected energy resolution
        self._resolution = 1 / (self._norm / self._fs)**0.5


    def _filter_traces(self, traces):
        """
        Method for carrying out the optimal filter in time domain via
        correlation.

        """

        alltraces = traces[:, self._tchan]

        # apply the FIR filter to each trace
        filts = np.array(
            [correlate(
                trace, self._phi, mode="same",
            ) / self._norm for trace in alltraces]
        )

        # set the filtered values to zero near the edges, so as not
        # to use the padded values in the analysis also so that the
        # traces that will be saved will be equal to the tracelength
        cut_len = np.max([len(self._phi), self._tracelength])

        filts[:, :cut_len//2] = 0
        filts[:, -(cut_len//2) + (cut_len + 1) % 2:] = 0

        return filts


    @staticmethod
    def _smart_trigger(trace, threshold_on, threshold_off,
                       mergewindow):
        """
        Method for carrying out a triggering algorithm that supports
        different turn-on and turn-off thresholds.

        """

        turn_on = (trace > threshold_on)
        turn_off = (trace > threshold_off)

        ind1 = 0
        ind_list = []

        searching = True

        while searching:

            ind_on = np.argmax(turn_on[ind1:]) + ind1
            ind_off = np.argmin(turn_off[ind_on:]) + ind_on

            if ind_on == ind_off:
                searching = False # for verbosity
                break

            ind_list.append([ind_on, ind_off])
            ind1 = ind_off

        if len(ind_list)==0:
            return []

        ind_array = np.vstack(ind_list)

        if mergewindow is not None:
            ind_array_flat = ind_array.flatten()
            arr = (ind_array_flat[1:] - ind_array_flat[:-1])[1::2] < mergewindow

            inds_keep = [0]
            for ii, b in enumerate(arr):
                if ~b:
                    inds_keep.extend([2 * ii + 1, 2 * ii + 2])
            inds_keep.append(-1)

            ind_array_out = ind_array_flat[inds_keep].reshape(len(inds_keep)//2, 2)

            return ind_array_out

        return ind_array


    def acquire_pulses(self, template, psd, threshold_on, tchan, threshold_off=None, mergewindow=None):
        """
        Method to carry out the offline triggering algorithm based on
        the OF formalism in time domain. Only trigeers on one specified
        channel.

        Parameters
        ----------
        template : ndarray
            The amplitude-normalized signal template in time domain.
        psd : ndarray
            The two-sided power spectral density describing the noise
            environment, to be used with the OF.
        threshold_on : float
            The trigger activation threshold to set, in units of
            number of expected baseline resolution, e.g. 10 corresponds
            to a 10-sigma threshold. If positive, it is assumed that
            events with amplitudes above this value will be extracted.
            If negative, then events with amplitudes below this value
            will be extracted, i.e. events will be assumed to be
            negative going. The section of data that is marked above
            threshold until the data goes below `threshold_off`.
        tchan : int
            The channel, designated by array index, to set a threshold
            on and extract events with amplitudes above the threshold.
        threshold_off : float, optional
            The trigger deactivation threshold to set, in units of
            number of expected baseline resolution, e.g. 10 corresponds
            to a 10-sigma threshold. If not specified, defaults to
            `threshold_on - 2`, unless `threshold_on < 5`. In this
            scenario, `threshold_off` is the smaller of 3 and
            `threshold_on`.
        mergewindow : int, NoneType, optional
            Window within which to merge triggers, in units of number of
            time bins. Defaults to no merging. It is not recommended to
            set this to above half of a trace length, as substantial
            dead time may be accrued.
        
        """

        self._template = template
        self._psd = psd
        self._nthreshold_on = threshold_on

        posthreshold = True if self._nthreshold_on > 0 else False

        if threshold_off is None:
            sign = 1 if posthreshold else -1
            if abs(self._nthreshold_on) > 5:
                self._nthreshold_off = threshold_on - sign * 2
            elif abs(self._nthreshold_on) > 3:
                self._nthreshold_off = 3 * sign * self._resolution
            else: self._nthreshold_off = threshold_on
        else:
            self._nthreshold_off = threshold_off

        self._tchan = tchan

        self._initialize_filter()
        self._threshold_on = self._nthreshold_on * self._resolution
        self._threshold_off = self._nthreshold_off * self._resolution


        savename = "trigger_" + self._start.strftime("%Y%m%d_%H%M%S")
        seriesnumber = int(self._start.strftime("%y%m%d%H%M%S"))

        filelist = sorted(glob(f"{self._contdatadir}/*.h5"))

        evtinds_list = []
        triginds_list = []
        evtamps_list = []
        traces_list = []
        parentsn_list = []
        parenten_list = []
        epochtime_list = []

        evt_counter = 0
        dumpnum = 1
        basenevents = 0


        for filename in filelist:

            FR = Reader(filename)
            data, metadata = FR.get_data(include_metadata=True)
            parentsn = metadata['parentseriesnumber'][0]
            parenten = metadata['parenteventnumber'][0]
            epochtime_start = metadata['eventtime'][0]
            
            filtered = self._filter_traces(data)

            for kk, filt in enumerate(filtered):

                if posthreshold:
                    ranges = EventBuilder._smart_trigger(
                        filt, self._threshold_on, self._threshold_off, mergewindow,
                    )
                else:
                    ranges = EventBuilder._smart_trigger(
                        -filt, -self._threshold_on, -self._threshold_off, mergewindow,
                    )

                if len(ranges)==0:
                    break

                for ind0, ind1 in zip(ranges[:, 0], ranges[:, 1]):
                    indmax = ind0 + np.argmax(filt[ind0:ind1])
                    evtinds_list.append([indmax - self._tracelength//2])
                    triginds_list.append([indmax])
                    evtamps_list.append([filt[indmax]])
                    trace_save_start = indmax - self._tracelength//2
                    trace_save_end = indmax + self._tracelength//2
                    traces_list.append(
                        data[[kk], :, trace_save_start:trace_save_end]
                    )

                evt_counter += len(ranges)

                parentsn_list.append(parentsn * np.ones(len(ranges)))
                parenten_list.append(parenten * np.ones(len(ranges)))
                epochtime_list.append(epochtime_start * np.ones(len(ranges)))

                if evt_counter >= self._maxevtsperdump:

                    evtinds = np.concatenate(evtinds_list)
                    triginds = np.concatenate(triginds_list)
                    evtamps = np.concatenate(evtamps_list)
                    traces = np.concatenate(traces_list)
                    parentsns = np.concatenate(parentsn_list)
                    parentens = np.concatenate(parenten_list)
                    epochtimes = np.concatenate(epochtime_list)

                    del evtinds_list
                    del triginds_list
                    del evtamps_list
                    del traces_list
                    del parentsn_list
                    del parenten_list
                    del epochtime_list

                    for ii in range(len(evtinds) // self._maxevtsperdump):

                        nevents = len(evtinds[:self._maxevtsperdump])
                        FW = Writer(
                            f"{self._savepath}{savename}_F{dumpnum:04d}.h5",
                        )
                        FW.write_data(
                            data=traces[:self._maxevtsperdump],
                            eventindex=evtinds[:self._maxevtsperdump],
                            eventnumber=np.arange(nevents) + basenevents,
                            eventtime=(
                                epochtimes[:self._maxevtsperdump]
                            ) + (
                                evtinds[:self._maxevtsperdump] / self._fs
                            ),
                            seriesnumber=[seriesnumber] * nevents,
                            dumpnumber=[dumpnum] * nevents,
                            triggertime=(
                                epochtimes[:self._maxevtsperdump]
                            ) + (
                                triginds[:self._maxevtsperdump] / self._fs
                            ),
                            triggertype=np.ones(nevents, dtype=int),
                            triggeramp=evtamps[:self._maxevtsperdump],
                            parentseriesnumber=parentsns[:self._maxevtsperdump],
                            parenteventnumber=parentens[:self._maxevtsperdump],
                            datashape=traces[:self._maxevtsperdump].shape,
                            fs=self._fs,
                            channels=metadata['channels'],
                            comment=f"""trigger:
                            turn-on threshold: {self._nthreshold_on}sigma,
                            turn-off threshold: {self._nthreshold_off}sigma""",
                            template=self._template,
                            psd=self._psd,
                        )
                        dumpnum += 1
                        basenevents += nevents

                        evtinds = evtinds[self._maxevtsperdump:]
                        triginds = triginds[self._maxevtsperdump:]
                        evtamps = evtamps[self._maxevtsperdump:]
                        traces = traces[self._maxevtsperdump:]
                        parentsns = parentsns[self._maxevtsperdump:]
                        parentens = parentens[self._maxevtsperdump:]
                        epochtimes = epochtimes[self._maxevtsperdump:]

                    if len(evtinds) > 0:
                        evtinds_list = [evtinds]
                        triginds_list = [triginds]
                        evtamps_list = [evtamps]
                        traces_list = [traces]
                        parentsn_list = [parentsns]
                        parenten_list = [parentens]
                        epochtime_list = [epochtimes]
                        evt_counter = len(evtinds)
                    else:
                        evtinds_list = []
                        triginds_list = []
                        evtamps_list = []
                        traces_list = []
                        parentsn_list = []
                        parenten_list = []
                        epochtime_list = []
                        evt_counter = 0

        # clean up the remaining events
        if evt_counter > 0:

            evtinds = np.concatenate(evtinds_list)
            triginds = np.concatenate(triginds_list)
            evtamps = np.concatenate(evtamps_list)
            traces = np.concatenate(traces_list)
            parentsns = np.concatenate(parentsn_list)
            parentens = np.concatenate(parenten_list)
            epochtimes = np.concatenate(epochtime_list)

            del evtinds_list
            del triginds_list
            del evtamps_list
            del traces_list
            del parentsn_list
            del parenten_list
            del epochtime_list

            dumps_left = len(evtinds) / self._maxevtsperdump

            for ii in range(np.ceil(dumps_left).astype(int)):

                nevents = len(evtinds[:self._maxevtsperdump])
                FW = Writer(f"{self._savepath}{savename}_F{dumpnum:04d}.h5")
                FW.write_data(
                    data=traces[:self._maxevtsperdump],
                    eventindex=evtinds[:self._maxevtsperdump],
                    eventnumber=np.arange(nevents) + basenevents,
                    eventtime=(
                        epochtimes[:self._maxevtsperdump]
                    ) + (
                        evtinds[:self._maxevtsperdump] / self._fs
                    ),
                    seriesnumber=[seriesnumber] * nevents,
                    dumpnumber=[dumpnum] * nevents,
                    triggertime=(
                        epochtimes[:self._maxevtsperdump]
                    ) + (
                        triginds[:self._maxevtsperdump] / self._fs
                    ),
                    triggertype=np.ones(nevents, dtype=int),
                    triggeramp=evtamps[:self._maxevtsperdump],
                    parentseriesnumber=parentsns[:self._maxevtsperdump],
                    parenteventnumber=parentens[:self._maxevtsperdump],
                    datashape=traces[:self._maxevtsperdump].shape,
                    fs=self._fs,
                    channels=metadata['channels'],
                    comment=f"""trigger:
                    turn-on threshold: {self._nthreshold_on}sigma,
                    turn-off threshold: {self._nthreshold_off}sigma""",
                    template=self._template,
                    psd=self._psd,
                )
                dumpnum += 1
                basenevents += nevents

                if ii + 1 != np.ceil(dumps_left).astype(int):
                    evtinds = evtinds[self._maxevtsperdump:]
                    triginds = triginds[self._maxevtsperdump:]
                    evtamps = evtamps[self._maxevtsperdump:]
                    traces = traces[self._maxevtsperdump:]
                    parentsns = parentsns[self._maxevtsperdump:]
                    parentens = parentens[self._maxevtsperdump:]
                    epochtimes = epochtimes[self._maxevtsperdump:]




import time
import os
import numpy as np

from moku.instruments import Datalogger


__all__ = [
    "LogData",
]


class LogData(object):
    """
    Class for managing the Datalogger object from the Moku package
    to take data with various settings safely.

    Attributes
    ----------
    DL : moku.instruments.Datalogger
        The wrapped Datalogger object that interfaces with the Moku.

    """

    def __init__(self, ip_address, force_connect=False, fs=None,
                 max_duration_per_file=60,
                 acquisition_mode="Precision"):
        """
        Initialization of the LogData class.

        Parameters
        ----------
        ip_address : str
            The IP address of the Moku to connect to. If connecting via
            IPv6 (e.g. via USB-C), ensure to enclose the address with
            square brackets, e.g. "[your_ip_address]".
        force_connect : bool, optional
            Take ownership of the Moku even if it is being used by
            someone else. Default is False.
        fs : float, optional
            The digitization rate of the data to be taken in Hz. The
            allowed values are 10 to 10e6 Hz for the Moku:Pro for 1
            channel logging, where the max decreases to 5e6 Hz for 2
            channel logging and to 1.25e6 for 3 or 4 channel logging.
            For the Moku:Lab, the allowed values are 10 to 2.5e5 Hz.
            For the Moku:Go, the allowed values are 10 to 1e6 Hz for
            1 channel logging, where the maximum allowed value
            decreases to 500e3 Hz for 2 channel logging. Default is the
            lowest allowed maximum value for the given device (i.e.
            1.25e6 for Moku:Pro, 2.5e5 for Moku:Lab, and 500e3 for
            Moku:Go).
        max_duration_per_file : float, optional
            The maximum amount of seconds to save to a single file.
            Avoids creating very large single files. Default is 60
            seconds.
        acquisition_mode : str, optional
            Changes acquisition mode between 'Normal' and 'Precision'.
            Precision mode is also known as decimation, it samples at
            the full rate and averages excess data points to improve
            precision. Normal mode works by direct down sampling,
            throwing away extra data points. Default is 'Precision'.

        """

        self.DL = Datalogger(
            ip_address, force_connect=force_connect, session_trust_env=False,
        )
        self.DL.set_acquisition_mode(acquisition_mode)
        self._max_dur_per_file = max_duration_per_file

        self._device = self.DL.describe()['hardware']
        if self._device not in ['Moku:Pro', 'Moku:Lab', 'Moku:Go']:
            raise ValueError(
                "Unrecognized device, is not a Moku:Go, Moku:Lab, or Moku:Pro."
            )

        if fs is None:
            if self._device == "Moku:Pro":
                self._fs = 1.25e6
            elif self._device == "Moku:Lab":
                self._fs = 2.5e5
            elif self._device == "Moku:Go":
                self._fs = 500e3
        else:
            self._fs = fs

    def __enter__(self):
        """Returns self when entered via with."""
        return self

    def __exit__(self, type, value, traceback):
        """
        Always run relinquish ownership and turn off inputs and outputs
        when exiting.

        """

        if self._device == "Moku:Pro":
            chan_list = [1, 2, 3, 4]
        elif self._device in ["Moku:Lab", "Moku:Go"]:
            chan_list = [1, 2]

        for chan in chan_list:
            self.DL.enable_input(chan, False) # disable inputs
            self.DL.generate_waveform(chan, 'Off') # disable outputs

        self.DL.relinquish_ownership()

    def set_input_channels(self, channels, impedance="1MOhm", coupling="DC",
                           vrange=None):
        """
        Set which channels to save data from and which settings to use.

        Parameters
        ----------
        channels : int, list of int
            Which input channels to log data from. Can be a single
            value for one channel, or a list of the channels. For the
            Moku:Pro, valid channel numbers are 1, 2, 3, and 4. For the
            Moku:Lab and Moku:Go, valid channel numbers are 1 and 2.
        impedance : str, list of str, optional
            The output impedance to use for each channel. For the
            Moku:Pro and Moku:Lab, options are '1MOhm' and '50Ohm'. For
            the Moku:Go, the only option is '1MOhm'. Can pass a list of
            the same length as channels if different impedances are
            desired for a Moku:Pro. Default is '1MOhm' for all channels.
        coupling : str, list of str, optional
            The coupling to use for each channel. Options are 'DC' and
            'AC'. Can pass a list of the same length as channels if
            different couplings are desired. Default is 'DC' for all
            channels.
        vrange : str, list of str, optional
            The voltage range to use for each channel. For the
            Moku:Pro, options are '400mVpp', '4Vpp' and '40Vpp'. For
            the Moku:Lab, the options are '1Vpp' and '10Vpp'. For the
            Moku:Go, options are '10Vpp' and '50Vpp'. Can pass a list
            of the same length as channels if different voltage ranges
            are desired for specific channels. Default is the smallest
            allowed value for the device for all channels (i.e.
            '400mVpp' for the Moku:Pro, '1Vpp' for the Moku:Lab, and
            '10Vpp' for the Moku:Go).

        """

        if self._device == "Moku:Pro":
            chan_list = [1, 2, 3, 4]
            vrange = '400mVpp' if vrange is None else vrange
        elif self._device == "Moku:Lab":
            chan_list = [1, 2]
            vrange = '1Vpp' if vrange is None else vrange
        elif self._device == "Moku:Go":
            chan_list = [1, 2]
            vrange = '10Vpp' if vrange is None else vrange

        if np.isscalar(channels):
            channels = [channels]
        if np.isscalar(impedance):
            impedance = [impedance] * len(channels)
        if np.isscalar(coupling):
            coupling = [coupling] * len(channels)
        if np.isscalar(vrange):
            vrange = [vrange] * len(channels)

        disable_chans = [chan not in channels for chan in chan_list]
        for chan, disable in zip(chan_list, disable_chans):
            if disable:
                self.DL.enable_input(chan, False)
            else:
                ind = channels.index(chan)
                self.DL.set_frontend(
                    channel=chan,
                    impedance=impedance[ind],
                    coupling=coupling[ind],
                    range=vrange[ind],
                )

    @staticmethod
    def sine_settings(amplitude=1, frequency=10000, offset=0, phase=0,
                      symmetry=50):
        """
        Method to return a dictionary with the valid sine wave
        settings to be passed to `self.set_output_channel`.

        Parameters
        ----------
        amplitude : float, optional
            Waveform peak-to-peak amplitude in Volts. Allowed values
            are 2e-3 to 10 for the Moku:Go, 2e-3 to 4 for the
            Moku:Lab, and 1e-3 to 10 for the Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 1 V.
        frequency : float, optional
            Waveform frequency in Hz. Allowed values are 1e-3 to 20e6
            for the Moku:Go, 1e-3 to 100e6 for the Moku:Lab, and 1e-3
            to 500e6 for the Moku:Pro. Default is 10000 (1e4) Hz.
        offset : float, optional
            DC offset applied to the waveform in V. Allowed values are
            -5 to 5 for Moku:Go, Moku:Lab, and Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 0.
        phase : float, optional
            Waveform phase offset in degrees. Allowed values are 0 to
            360 for Moku:Go, Moku:Lab, and Moku:Pro. Default is 0.
        symmetry : float, optional
            Fraction of the cycle rising in %. Allowed values are 0.0
            to 100.0 for Moku:Go, Moku:Lab, and Moku:Pro. Default is
            50.

        Returns
        -------
        settings_dict : dict

        """

        return {
            'amplitude': amplitude,
            'frequency': frequency,
            'offset': offset,
            'phase': phase,
            'symmetry': symmetry,
        }

    @staticmethod
    def square_settings(amplitude=1, frequency=10000, offset=0, phase=0,
                        duty=50, symmetry=50):
        """
        Method to return a dictionary with the valid square wave
        settings to be passed to `self.set_output_channel`.

        Parameters
        ----------
        amplitude : float, optional
            Waveform peak-to-peak amplitude in Volts. Allowed values
            are 2e-3 to 10 for the Moku:Go, 2e-3 to 4 for the
            Moku:Lab, and 1e-3 to 10 for the Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 1 V.
        frequency : float, optional
            Waveform frequency in Hz. Allowed values are 1e-3 to 20e6
            for the Moku:Go, 1e-3 to 100e6 for the Moku:Lab, and 1e-3
            to 500e6 for the Moku:Pro. Default is 10000 (1e4) Hz.
        offset : float, optional
            DC offset applied to the waveform in V. Allowed values are
            -5 to 5 for Moku:Go, Moku:Lab, and Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 0.
        phase : float, optional
            Waveform phase offset in degrees. Allowed values are 0 to
            360 for Moku:Go, Moku:Lab, and Moku:Pro. Default is 0.
        duty : float, optional
            Duty cycle as percentage in %. Allowed values are 0 to 100
            for Moku:Go and Moku:Pro. Default is 0.
        symmetry : float, optional
            Fraction of the cycle rising in %. Allowed values are 0.0
            to 100.0 for Moku:Go, Moku:Lab, and Moku:Pro. Default is
            50.

        Returns
        -------
        settings_dict : dict
            A dictionary containing all of the required settings
            needed to set up a square wave output.

        """

        return {
            'amplitude': amplitude,
            'frequency': frequency,
            'offset': offset,
            'phase': phase,
            'duty': duty,
            'symmetry': symmetry,
        }

    @staticmethod
    def ramp_settings(amplitude=1, frequency=10000, offset=0, phase=0,
                      symmetry=50):
        """
        Method to return a dictionary with the valid ramp wave
        settings to be passed to `self.set_output_channel`.

        Parameters
        ----------
        amplitude : float, optional
            Waveform peak-to-peak amplitude in Volts. Allowed values
            are 2e-3 to 10 for the Moku:Go, 2e-3 to 4 for the
            Moku:Lab, and 1e-3 to 10 for the Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 1 V.
        frequency : float, optional
            Waveform frequency in Hz. Allowed values are 1e-3 to 20e6
            for the Moku:Go, 1e-3 to 100e6 for the Moku:Lab, and 1e-3
            to 500e6 for the Moku:Pro. Default is 10000 (1e4) Hz.
        offset : float, optional
            DC offset applied to the waveform in V. Allowed values are
            -5 to 5 for Moku:Go, Moku:Lab, and Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 0.
        phase : float, optional
            Waveform phase offset in degrees. Allowed values are 0 to
            360 for Moku:Go, Moku:Lab, and Moku:Pro. Default is 0.
        symmetry : float, optional
            Fraction of the cycle rising in %. Allowed values are 0.0
            to 100.0 for Moku:Go, Moku:Lab, and Moku:Pro. Default is
            50.

        Returns
        -------
        settings_dict : dict
            A dictionary containing all of the required settings
            needed to set up a ramp wave output.

        """

        return {
            'amplitude': amplitude,
            'frequency': frequency,
            'offset': offset,
            'phase': phase,
            'symmetry': symmetry,
        }

    @staticmethod
    def pulse_settings(amplitude=1, frequency=10000, offset=0, phase=0,
                       symmetry=50, edge_time=0, pulse_width=0):
        """
        Method to return a dictionary with the valid pulse train
        settings to be passed to `self.set_output_channel`.

        Parameters
        ----------
        amplitude : float, optional
            Waveform peak-to-peak amplitude in Volts. Allowed values
            are 2e-3 to 10 for the Moku:Go, 2e-3 to 4 for the
            Moku:Lab, and 1e-3 to 10 for the Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 1 V.
        frequency : float, optional
            Waveform frequency in Hz. Allowed values are 1e-3 to 20e6
            for the Moku:Go, 1e-3 to 100e6 for the Moku:Lab, and 1e-3
            to 500e6 for the Moku:Pro. Default is 10000 (1e4) Hz.
        offset : float, optional
            DC offset applied to the waveform in V. Allowed values are
            -5 to 5 for Moku:Go, Moku:Lab, and Moku:Pro (For Moku:Pro,
            the output voltage is limited to between -1V and 1V above
            1MHz). Default is 0.
        phase : float, optional
            Waveform phase offset in degrees. Allowed values are 0 to
            360 for Moku:Go, Moku:Lab, and Moku:Pro. Default is 0.
        symmetry : float, optional
            Fraction of the cycle rising in %. Allowed values are 0.0
            to 100.0 for Moku:Go, Moku:Lab, and Moku:Pro. Default is
            50.
        edge_time : float, optional
            Edge-time of the waveform in s. Allowed values are 16e-9 to
            pulse_width for the Moku:Go, 4e-9 to pulse_width for the
            Moku:Lab, and 2e-9 to pulse_width for the Moku:Pro. Default
            is 0 (which sets it to the shortest allowed time).
        pulse_width : float, optional
            Pulse width of the waveform in s. Allowed values are 16e-9
            to waveform period for the Moku:Go, 4e-9 to waveform period
            for the Moku:Lab and 2e-9 to waveform period for the
            Moku:Pro. Default is 0 (which sets it to the shortest
            allowed time).

        Returns
        -------
        settings_dict : dict
            A dictionary containing all of the required settings
            needed to set up a pulse train output.

        """

        return {
            'amplitude': amplitude,
            'frequency': frequency,
            'offset': offset,
            'phase': phase,
            'symmetry': symmetry,
            'edge_time': edge_time,
            'pulse_width': pulse_width,
        }

    @staticmethod
    def dc_settings(dc_level=0):
        """
        Method to return a dictionary with the valid DC value
        settings to be passed to `self.set_output_channel`.

        Parameters
        ----------
        dc_level : float, optional
            The DC level of the output of the channel in V.
            Default is 0.

        Returns
        -------
        settings_dict : dict
            A dictionary containing all of the required settings
            needed to set up a DC output.

        """

        return {
            'dc_level': dc_level,
        }

    def set_output_channel(self, channel, waveformtype, load="HiZ",
                           **settings):
        """
        Method to turn on an output channel and generate the speicified
        waveform. Use the various self.*_settings() methods to
        generate the various valid settings. Can only set one channel
        at a time.

        Parameters
        ----------
        channel : int
            The output channel to generate a waveform on. With a
            Moku:Pro, this must be an integer of 1, 2, 3 or 4. With a
            Moku:Go or Moku:Lab, this must be an integer of 1 or 2.
        waveformtype : str
            The waveform type to generate, must be one of 'Sine',
            'Square', 'Ramp', 'Pulse', 'DC'. Can also be set to 'Off'
            to turn the channel off.
        termination : str, optional
            The waveform termination to use for the output channel.
            For a Moku:Pro or Moku:Lab, this must be one of 'HiZ' or
            '50Ohm'. For a Moku:Go, this can only be 'HiZ'. Default
            is 'HiZ'.
        settings : dict
            The dictionary containing all of the settings needed for
            the specified waveform type.

        """

        if self._device in ["Moku:Pro", "Moku:Lab"]:
            self.DL.set_output_termination(channel, load)
        self.DL.generate_waveform(
            channel,
            waveformtype,
            **settings,
        )


    def log_data(self, duration, savepath='./', comments='',
                 file_name_prefix=''):
        """
        Method to log data after setting up the desired configuration.
        Saves a LI file to the specified save path.

        Parameters
        ----------
        duration : integer
            The total number of seconds to log data with the current
            configuration. Must be an integer.
        savepath : str, optional
            The absolute path on to save the LI file after it has been
            logged by the Moku. Default is the current working
            directory.
        comments : str, optional
            Any comments that should be added to the metadata of the
            file to be saved. Default is an empty string.
        file_name_prefix : str, optional
            Prefix to use in the filename. Default is set by the Moku
            as "MokuDataLoggerData".

        """

        self.DL.set_samplerate(self._fs)

        filenames = []
        nfiles = np.int32(np.ceil(duration / self._max_dur_per_file))

        for ii in range(nfiles):
            if (ii + 1) * self._max_dur_per_file > duration:
                this_file_duration = duration % self._max_dur_per_file
            else:
                this_file_duration = self._max_dur_per_file

            logfile = self.DL.start_logging(
                duration=this_file_duration,
                comments=comments,
                file_name_prefix=file_name_prefix,
            )
            # Track progress percentage of the data logging session
            is_logging = True
            while is_logging:
                # Wait for the logging session to progress by sleeping 0.5sec
                time.sleep(0.5)
                # Get current progress percentage and check if it's complete
                progress = self.DL.logging_progress()
                is_logging = not progress['complete']

            filenames.append(logfile['file_name'])

        if self._device == "Moku:Pro":
            target = "ssd"
        elif self._device == "Moku:Lab":
            target = "media"
        elif self._device == "Moku:Go":
            target = "persist"

        for fname in filenames:
            self.DL.download(
                target,
                fname,
                os.path.abspath(savepath + os.sep + fname),
            )



import serial

__all__ = [
    "HarvardPowerSupply",
]

class HarvardPowerSupply(object):
    """
    Class for controlling the Harvard Power Supply at the SLAC
    SPLENDOR lab.

    Notes
    -----
    The documentation of the power supply can be found here
    https://wiki.harvard.edu/confluence/display/ESHOP/TT075%3A+Stanford+HEMT+supply

    The serial commands are reprinted below in case the above link no
    longer works, or the use has no access.

    Commands:

    S <arg>;  selects active slot.  Valid arguments are 1 and 2.
        Power-on default is 1.
    C <arg>; selects active HEMT channel.  Valid arguments are 1 and 2.
        Power-on default is 1.
    D <arg>; sets drain voltage of active HEMT channel in millivolts.
        Valid arguments are floats in the range of 0.0 to 4990.0.
    d; reads back drain voltage of selected HEMT channel.  Note that
        this is mostly for diagnostics, as the "D" command is at least
        as accurate as the "d" command.
    r; reads back remote voltage (Vds+ - Vds-) of selected HEMT channel.
        Mostly for diagnostics unless you anticipate significant
        resistance on the Vds line to the HEMT.
    G <arg>; sets the Vgs in millivolts. Valid arguments are in the
        range of -5000.0 mV to +5000.0 mV.   Power-on default sets Vgs
        to 0.0V.
    g; reads back the previously set Gate voltage.  This is a
        firmware-only operation, simply returning the previously set
        Vgs, rather than measuring it.
    N; turns oN the Vds driver of both HEMTs of the active slot.
        Power-on default has the drivers turned off .
    F; turns ofF the Vds driver of both HEMTs of the active slot.
    i; returns the drain current of the selected HEMT, in mA.

    The preferred protocol is to use "D" to set the correct Vds,
    then use "N" to enable the output drivers.

    """

    def __init__(self, port):
        """
        Initialization of the HarvardPowerSupply class. Establishes
        a connection to a specified port.

        Parameters
        ----------
        port : str
            The port to connect to that the HarvardPowerSupply is on,
            e.g. "COM4".

        """

        self._HPS = serial.Serial()
        self._HPS.port = port
        self._HPS.open()
        self.nslots = 2
        self.nhemts = 2

    def choose_slot(self, slot):
        """
        Choose which slot to control.

        Parameters
        ----------
        slot : int
            The slot to connect to, should be 1 or 2.

        Returns
        -------
        slot_val : int
            The value of the slot that was set.

        """

        self._HPS.write(f"S{slot};".encode())
        slot_str = self._HPS.read_until(b"!")
        self._HPS.write(b'N;')
        self._HPS.read_until(b'!')
        return int(slot_str[1:-1])

    def choose_hemt(self, hemt):
        """
        Choose which HEMT to control on the specified slot.

        Parameters
        ----------
        hemt : int
            The HEMT to connect to on the previously specified slot,
            should be 1 or 2.

        Returns
        -------
        hemt_val : int
            The value of the HEMT that was set.

        """

        self._HPS.write(f'C{hemt};'.encode())
        return int(self._HPS.read_until(b'!')[1:-1])

    def set_gate_voltage(self, value):
        """
        With HEMT and slot chosen, set the gate voltage on the HEMT.

        Parameters
        ----------
        value : float
            The gate voltage to be set for the HEMT, in mV. Allowed
            values are -5000.0 to 5000.0, power-on defaults to 0.

        Returns
        -------
        voltage_val : float
            The value of the gate voltage that was set, in mV.

        """

        self._HPS.write(f"G{value};".encode())
        return float(self._HPS.read_until(b'!')[1:-1])

    def read_gate_voltage(self):
        """
        With HEMT and slot chosen, read the gate voltage on the HEMT.

        Returns
        -------
        voltage_val : float
            The value of the gate voltage that was set, in mV.

        """

        self._HPS.write(b"g;")
        return float(self._HPS.read_until(b'!')[1:-1])

    def set_drain_voltage(self, value):
        """
        With HEMT and slot chosen, set the drain voltage on the HEMT.

        Parameters
        ----------
        value : float
            The drain voltage to be set for the HEMT, in mV. Allowed
            values are 0.0 to 4990.0.

        Returns
        -------
        voltage_val : float
            The value of the gate voltage that was set, in mV.

        """

        self._HPS.write(f"D{value};".encode())
        return float(self._HPS.read_until(b'!')[1:-1])
    
    def read_drain_voltage(self):
        """
        With HEMT and slot chosen, read the drain voltage on the HEMT.

        Returns
        -------
        voltage_val : float
            The value of the drain voltage that was set, in mV.

        """

        self._HPS.write(b"d;")
        return float(self._HPS.read_until(b'!')[1:-1])

    def read_remote_voltage(self):
        """
        With HEMT and slot chosen, read the remote voltage
        (Vds+ - Vds-) on the HEMT.

        Returns
        -------
        voltage_val : float
            The value of the remote voltage in mV.

        """

        self._HPS.write(b"d;")
        return float(self._HPS.read_until(b'!')[1:-1])

    def read_drain_current(self):
        """
        With HEMT and slot chosen, read the drain current on the HEMT.

        Returns
        -------
        current_val : float
            The value of the drain current in mA.

        """

        self._HPS.write(b"i;")
        return float(self._HPS.read_until(b'!')[1:-1])
    
    def zero_all(self):
        """
        Function to zero all voltages, good to run at beginning and end
        of running to ensure voltages are zero and the environment is
        known.

        Returns
        -------
        None

        """

        for slot in range(1, self.nslots + 1):
            self.choose_slot(slot)
            for hemt in range(1, self.nhemts + 1):
                self.choose_hemt(hemt)
                self.set_gate_voltage(0)
                self.set_drain_voltage(0)

    def close(self):
        """
        Close connection to the port to free it for other users.

        Returns
        -------
        None

        """

        self._HPS.close()



import numpy as np
import os
import time

from moku.instruments import Oscilloscope as Osc


__all__ = [
    "Oscilloscope",
]


class Oscilloscope(object):
    """
    Class for managing the Oscilloscope object from the Moku package
    to take data with various settings safely.

    Attributes
    ----------
    Osc : moku.instruments.Oscilloscope
        The wrapped Oscilloscope object that interfaces with the Moku.

    """

    def __init__(self, ip_address, force_connect=False,
                 acquisition_mode="Precision"):
        """
        Initialization of the Oscilloscope class.

        Parameters
        ----------
        ip_address : str
            The IP address of the Moku to connect to. If connecting via
            IPv6 (e.g. via USB-C), ensure to enclose the address with
            square brackets, e.g. "[your_ip_address]".
        force_connect : bool, optional
            Take ownership of the Moku even if it is being used by
            someone else. Default is False.
        acquisition_mode : str, optional
            Changes acquisition mode between 'Normal' and 'Precision'.
            Precision mode is also known as decimation, it samples at
            the full rate and averages excess data points to improve
            precision. Normal mode works by direct down sampling,
            throwing away extra data points. Default is 'Precision'.

        """

        self.Osc = Osc(
            ip_address, force_connect=force_connect, session_trust_env=False,
        )
        self.Osc.set_acquisition_mode(acquisition_mode)

        self._device = self.Osc.describe()['hardware']
        if self._device not in ['Moku:Pro', 'Moku:Lab', 'Moku:Go']:
            raise ValueError(
                "Unrecognized device, is not a Moku:Go, Moku:Lab, or Moku:Pro."
            )

    def __enter__(self):
        """Returns self when entered via with."""
        return self

    def __exit__(self, type, value, traceback):
        """
        Always run relinquish ownership and turn off inputs and outputs
        when exiting.

        """

        if self._device == "Moku:Pro":
            chan_list = [1, 2, 3, 4]
        elif self._device in ["Moku:Lab", "Moku:Go"]:
            chan_list = [1, 2]

        for chan in chan_list:
            self.Osc.generate_waveform(chan, 'Off') # disable outputs
            self.Osc.disable_input(chan) # disable inputs

        self.Osc.relinquish_ownership()

    def set_input_channels(self, channels, impedance="1MOhm", coupling="DC",
                           vrange=None):
        """
        Set which channels to save data from and which settings to use.

        Parameters
        ----------
        channels : int, list of int
            Which input channels to log data from. Can be a single
            value for one channel, or a list of the channels. For the
            Moku:Pro, valid channel numbers are 1, 2, 3, and 4. For the
            Moku:Lab and Moku:Go, valid channel numbers are 1 and 2.
        impedance : str, list of str, optional
            The output impedance to use for each channel. For the
            Moku:Pro and Moku:Lab, options are '1MOhm' and '50Ohm'. For
            the Moku:Go, the only option is '1MOhm'. Can pass a list of
            the same length as channels if different impedances are
            desired for a Moku:Pro. Default is '1MOhm' for all channels.
        coupling : str, list of str, optional
            The coupling to use for each channel. Options are 'DC' and
            'AC'. Can pass a list of the same length as channels if
            different couplings are desired. Default is 'DC' for all
            channels.
        vrange : str, list of str, optional
            The voltage range to use for each channel. For the
            Moku:Pro, options are '400mVpp', '4Vpp' and '40Vpp'. For
            the Moku:Lab, the options are '1Vpp' and '10Vpp'. For the
            Moku:Go, options are '10Vpp' and '50Vpp'. Can pass a list
            of the same length as channels if different voltage ranges
            are desired for specific channels. Default is the smallest
            allowed value for the device for all channels (i.e.
            '400mVpp' for the Moku:Pro, '1Vpp' for the Moku:Lab, and
            '10Vpp' for the Moku:Go).

        """

        if self._device == "Moku:Pro":
            chan_list = [1, 2, 3, 4]
            vrange = '400mVpp' if vrange is None else vrange
        elif self._device == "Moku:Lab":
            chan_list = [1, 2]
            vrange = '1Vpp' if vrange is None else vrange
        elif self._device == "Moku:Go":
            chan_list = [1, 2]
            vrange = '10Vpp' if vrange is None else vrange

        if np.isscalar(channels):
            channels = [channels]
        if np.isscalar(impedance):
            impedance = [impedance] * len(channels)
        if np.isscalar(coupling):
            coupling = [coupling] * len(channels)
        if np.isscalar(vrange):
            vrange = [vrange] * len(channels)

        disable_chans = [chan not in channels for chan in chan_list]
        for chan, disable in zip(chan_list, disable_chans):
            if disable:
                self.Osc.disable_input(chan)
            else:
                ind = channels.index(chan)
                self.Osc.set_frontend(
                    channel=chan,
                    impedance=impedance[ind],
                    coupling=coupling[ind],
                    range=vrange[ind],
                )

    def set_output_channels(self, channels, waveformtype='DC', load="1MOhm",
                            dc_level=0):
        """
        Method to turn on an output channel and generate the speicified
        waveform. Use the various self.*_settings() methods to
        generate the various valid settings. Can only set one channel
        at a time.

        Parameters
        ----------
        channel : int, list of int
            The output channel to generate a waveform on. With a
            Moku:Pro, this must be an integer of 1, 2, 3 or 4. With a
            Moku:Lab or Moku:Go, this must be an integer of 1 or 2.
        waveformtype : str, list of str, optional
            The waveform type to generate, only supports 'DC' waveforms
            at the moment. Can also be set to 'Off' to turn the channel
            off. Default is 'DC'.
        load : str, list of str, optional
            The load impedance to use for the output channel. For a
            Moku:Pro or Moku:Lab, this must be one of '1MOhm' or
            '50Ohm'. For a Moku:Go, this can only be '1MOhm'. Default
            is '1MOhm'.
        dc_level : float, list of float, optional
            The DC level of the output of the channel in V. Default is
            0.

        """

        if np.isscalar(channels):
            channels = [channels]
        if np.isscalar(waveformtype):
            waveformtype = [waveformtype] * len(channels)
        if np.isscalar(load):
            load = [load] * len(channels)
        if np.isscalar(dc_level):
            dc_level = [dc_level] * len(channels)

        for ind, chan in enumerate(channels):
            if self._device in ["Moku:Pro", "Moku:Lab"]:
                self.Osc.set_output_load(chan, load[ind])
            self.Osc.generate_waveform(
                channel=chan,
                type=waveformtype[ind],
                dc_level=dc_level[ind],
            )

    def get_data(self, sources, duration):
        """
        Method to get data after setting up the desired configuration.
        Returns a dictionary with the time domain values.

        Parameters
        ----------
        sources : str, list of str
            The sources which to read out in order of increasing
            channel number. Can be some combination of "Input1",
            "Input2", "Output1", "Output2" for the Moku:Go or Moku:Lab,
            with "Input3", "Input4", "Output3", and "Output4" also
            available for the Moku:Pro.
        duration : float
            The total time in seconds to collect data with the current
            configuration.

        Returns
        -------
        data : dict
            Dictionary with the time-domain data of the sources
            specified. The keys of the sources are labeled as
            "ch1", "ch2", etc. The digitization rate of the data is
            always 1024 / `duration`.

        """

        for ii, s in enumerate(sources):
            self.Osc.set_source(ii + 1, s)

        self.Osc.set_timebase(0, duration)
        time.sleep(duration)

        data = self.Osc.get_data()

        return data



from datetime import datetime
import itertools
import numpy as np
import os
import time
import warnings
import yaml

from splendaq.daq import LogData, Oscilloscope


__all__ = [
    "Sequencer",
]


class Sequencer(object):
    """
    Class for running a DC sequencer with various inputs read out,
    while changing various outputs by specified steps.

    Attributes
    ----------
    yaml_dict : dict
        A dictionary containing all of the settings to be used for
        taking data with the sequencer.

    """

    def __init__(self, yaml_file):
        """
        Initialization of the Sequencer given a YAML file which
        contains all of the settings to be used.

        Parameters
        ----------
        yaml_file : str
            The absolute path to the YAML file to be used to set up
            the sequencer for data taking.

        """

        with open(yaml_file, 'r') as f:
            self.yaml_dict = yaml.safe_load(f)
        self._savepath = self.yaml_dict['moku']['savepath']

    def run(self, verbose=True):
        """
        Start running the sequencer.

        Parameters
        ----------
        verbose : bool, optional
            If True, prints out information on the current point
            being run by the sequencer.

        """

        if self.yaml_dict['moku']['sequencer_mode']=='full':
            self._full_run(verbose)
        elif self.yaml_dict['moku']['sequencer_mode']=='quick':
            self._quick_run(verbose)
        else:
            raise ValueError("Unrecognized sequencer mode type.")

    def _init_run(self):
        """Initialize inputs and outputs based on the YAML file."""

        input_list = [
            int(k[5:]) for k in self.yaml_dict if (
                "input" in k and self.yaml_dict[k]['log']
            )
        ]

        input_dict = [
            d for k, d in self.yaml_dict.items() if (
                "input" in k and self.yaml_dict[k]['log']
            )
        ]

        input_settings = {
            "vrange": [d['vrange'] for d in input_dict],
            "impedance": [d['impedance'] for d in input_dict],
        }

        output_list = [
            int(k[6:]) for k in self.yaml_dict if (
                "output" in k and self.yaml_dict[k]['apply']
            )
        ]

        output_ranges = [
            np.linspace(
                val['vstart'],
                val['vend'],
                num=val['nstep'],
            ) for k, val in self.yaml_dict.items() if (
                "output" in k and self.yaml_dict[k]['apply']
            )
        ]

        return input_list, input_settings, output_list, output_ranges

    @staticmethod
    def _print_status(outputs, output_list):
        """Hidden method to print current bias points of outputs."""
        print(
            ', '.join(
                [
                    f"Output{b} = {a} V" for a, b in zip(
                        outputs,
                        output_list,
                    )
                ]
            )
        )

    def _full_run(self, verbose):
        """
        Running the 'full' sequencer mode, which calls the LogData
        class to save the time-domain traces to LI files.

        """

        input_list, input_settings, output_list, output_ranges = self._init_run()

        with LogData(
            self.yaml_dict['moku']['ip'],
            force_connect=self.yaml_dict['moku']['force_connect'],
            acquisition_mode=self.yaml_dict['moku']['acquisition_mode'],
            ) as LOG:

            LOG.set_input_channels(input_list, **input_settings)

            for outputs in itertools.product(*output_ranges):

                for ii in output_list:
                    dc_settings = LOG.dc_settings(
                        dc_level=outputs[output_list.index(ii)],
                    )
                    LOG.set_output_channel(ii, 'DC', **dc_settings)

                time.sleep(self.yaml_dict['moku']['time_between_points'])

                if verbose:
                    self._print_status(outputs, output_list)

                LOG.log_data(
                    self.yaml_dict['moku']['duration_per_point'],
                    savepath=self._savepath,
                    file_name_prefix="splendaq_iv",
                )

    def _quick_run(self, verbose):
        """
        Running the 'quick' sequencer mode, which calls the
        Oscilloscope class to simply save DC values at each bias point.

        """

        input_list, input_settings, output_list, output_ranges = self._init_run()
        product_len = np.multiply.reduce(
            [len(arr) for arr in output_ranges] + [1]
        )
        data_array = np.zeros(
            (product_len, len(input_list) + len(output_list))
        )
        columns = [
            f"Input{ii}" for ii in input_list
        ] + [
            f"Output{ii}" for ii in input_list
        ]

        with Oscilloscope(
            self.yaml_dict['moku']['ip'],
            force_connect=self.yaml_dict['moku']['force_connect'],
            acquisition_mode=self.yaml_dict['moku']['acquisition_mode'],
            ) as Osc:

            Osc.set_input_channels(input_list, **input_settings)

            for n, outputs in enumerate(itertools.product(*output_ranges)):

                Osc.set_output_channels(output_list, 'DC', dc_level=outputs)

                time.sleep(self.yaml_dict['moku']['time_between_points'])

                if verbose:
                    self._print_status(outputs, output_list)

                data = Osc.get_data(
                    [f"Input{ii}" for ii in input_list],
                    self.yaml_dict['moku']['duration_per_point'],
                )

                for ii in range(len(input_list)):
                    if 'nan' in data[f'ch{ii + 1}']:
                        warnings.warn(
                            """NaNs detected in data at this bias """
                            """point, tossing the NaNs"""
                        )
                        arr_float = np.asarray(
                            data[f'ch{ii + 1}'], dtype=float,
                        )
                        data_array[n, ii] = np.nanmean(arr_float)
                    else:
                        data_array[n, ii] = np.mean(data[f'ch{ii + 1}'])

                data_array[n, len(input_list):] = outputs

        date_str = datetime.now().strftime('%Y%m%d_%H%M%S')
        np.savetxt(
            f'{self._savepath}{os.sep}splendaq_iv_{date_str}.txt',
            data_array,
            delimiter=',',
            header=','.join(columns),
        )




import argparse
import os
import sys
import yaml

import splendaq
from splendaq.daq import Sequencer

def splendaq_cli():
    """
    SPLENDAQ command line interface.

    """

    parser = argparse.ArgumentParser(
        prog="splendaq", description="splendaq's command-line interface"
    )

    # global options
    parser.add_argument(
        "--version", action="store_true", help="Print splendaq version and exit",
    )

    subparsers = parser.add_subparsers()

    add_sequencer_parser(subparsers)

    args = parser.parse_args()

    if args.version:
        print(splendaq.__version__)
        sys.exit()

    args.func(args)


def add_sequencer_parser(subparsers):
    """
    Configure the DC sequencer command line interface.

    """

    parser_sequencer = subparsers.add_parser(
        "sequencer", description="Run a DC sequencer on the Moku, logging input channels",
    )

    parser_sequencer.add_argument(
        "yaml_file", help="YAML file containing sequencer setup.",
    )

    parser_sequencer.add_argument(
        "--noprint",
        action="store_false",
        help="""Do not print output DC voltage values.""",
    )

    parser_sequencer.set_defaults(func=sequencer_cli)

def sequencer_cli(args):
    """
    Sequencer command line interface.
    
    """

    SEQ = Sequencer(args.yaml_file)
    SEQ.run(args.noprint)
